unit cDIBPanel;

{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: cDIBPanl.PAS, released August 28, 2000.

The Initial Developer of the Original Code is Peter Morris (pete@stuckindoors.com),
Portions created by Peter Morris are Copyright (C) 2000 Peter Morris.
All Rights Reserved.

Purpose of file:
The base class for DIBContainers.  All DIB components must be parented by a DIBContainer
of some sort.

Contributor(s):
azzaazza69@hotmail.com - www.azzasoft.com
RiceBall <riceb@nether.net>

Last Modified: June 24, 2001

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org
or at http://www.stuckindoors.com/dib


Known Issues:
To be updated !
-----------------------------------------------------------------------------}
//Modifications
(*
Date:   December 22, 2000
By:     Azza
Change: WMPaint didn't set up the PaintRect area for a foreign device context
        (ie. anything other than itself).

Date:   May 2, 2001
By:     Peter Morris
Change: Added DIBBorder property

Date:   May 2, 2001
By:     Peter Morris
Change: Added BeforePaint and AfterPaint;

Date:   June 24, 2001
By:     RiceBall
Change: Overrode AlignControls to take into account DIBBorder

Date:   June 24, 2001
By:     Peter Morris
Change: Added a property to determine if the border is draw over or under its controls

Date:   August 24, 2001
By:     Peter Morris
Change: Changed BeforePaint and AfterPaint to DoBeforePaint and DoAfterPaint,
        added BeforePaint and AfterPaint events
*)


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, cDIB, JPeg, cDIBImageList, cDIBPalette, cDIBBorder;

type
  TBorderDrawPosition = (bdOverControls, bdUnderControls);
  TDIBPaintEvent = procedure(Sender: TObject; const DIB: TAbstractSuperDIB) of object;

  TCustomDIBContainer = class(TCustomPanel)
  private
    { Private declarations }
    FOnActiveControlChange : TNotifyEvent;
    FBeforePaint: TDIBPaintEvent;
    FAfterPaint: TDIBPaintEvent;
    FTabbing        : Boolean;
    FChangingFocus  : Boolean;
    FAlteredRect    : Boolean;
    FActiveControl  : TControl;
    FTabList        : TList;
    FChildDIB       : TWinDIB;
    FDIB            : TWinDIB;
    FDIBBorder      : TDIBBorder;
    FUpdateRect     : TRect;
    FPalette        : TDIBPalette;
    FBorderDrawPosition: TBorderDrawPosition;

    procedure CMFocusChanged(var Message : TCMFocusChanged); message CM_FocusChanged;
    procedure CNChar(var Message : TWMKey); message CN_Char;
    procedure CNKeyDown(var Message: TWMKeyDown); message CN_KeyDown;
    procedure CNKeyUp(var Message : TWMKeyUp); message CN_KeyUp;
    procedure WMEraseBkGnd(var Message : TMessage); message WM_EraseBkGnd;
    procedure WMGetDlgCode(var Message : TMessage); message WM_GetDlgCode;
    procedure WMPaint(Var Message : TWMPaint); message WM_PAINT;
    procedure SetBorderDrawPosition(const Value: TBorderDrawPosition);
  protected
    { Protected declarations }
    procedure AlignControls(AControl: TControl; var Rect: TRect);override;
    function  ChildWantsKey(var Message : TWMKey) : Boolean;
    procedure CreateParams(var Params : TCreateParams); override;
    procedure DoAfterPaint; virtual;
    procedure DoBeforePaint; virtual;
    procedure DoEnter; override;
    procedure DoExit; override;
    procedure DoTabFixups; virtual;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure PaintControls(DC: HDC; First: TControl);
    procedure PaintHandler(var Message: TWMPaint);
    procedure SetDIBBorder(const Value: TDIBBorder); virtual;
    procedure WndProc(var Message : TMessage); override;

    //Focus routines
    function DIBGetFirst : TControl;
    function DIBGetPrior : TControl;
    function DIBGetNext  : TControl;
    function DIBGetLast  : TControl;
    function DIBGetIndex(aControl : TControl) : Integer;

    function DIBSelectNext : Boolean;
    function DIBSelectPrior: Boolean;

    property AfterPaint: TDIBPaintEvent read FAfterPaint write FAfterPaint;
    property BeforePaint: TDIBPaintEvent read FBeforePaint write FBeforePaint;
    property BorderDrawPosition: TBorderDrawPosition
      read FBorderDrawPosition
      write SetBorderDrawPosition;
    property ChildDIB         : TWinDIB
      read FChildDIB;
    property DIBBorder        : TDIBBorder
      read FDIBBorder
      write SetDIBBorder;
    property OnActiveControlChange      : TNotifyEvent
      read FOnActiveControlChange
      write FOnActiveControlChange;

  public
    { Public declarations }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;

    //Focus routines
    function  DIBGetTabOrder(aControl : TControl) : TTabOrder;
    procedure DIBSetTabOrder(aControl : TControl; NewIndex : TTabOrder);

    function  DIBFindNext(aControl : TControl; aForward : Boolean) : TControl;
    procedure DIBFocusControl(aControl : TControl); virtual;

    //Std
    procedure Paint; Override;
    procedure SetBounds(aLeft, aTop, aWidth, aHeight : Integer); override;

    property ActiveControl    : TControl
      read FActiveControl
      write DIBFocusControl;
    property DIB              : TWinDIB
      read FDIB;
    property UpdateRect       : TRect
      read FUpdateRect;
  published
    { Published declarations }
  end;

  TDIBContainer = class(TCustomDIBContainer)
  private
  protected
  public
  published
    property AfterPaint;
    property Align;
    property Anchors;
    property AutoSize;
    property BeforePaint;
    property BiDiMode;
    property BorderDrawPosition;
    property Color;
    property Constraints;
    property Cursor;
    property DockSite;
    property DIBBorder;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property Height;
    property HelpContext;
    property Hint;
    property Left;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property ParentBiDiMode;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property Tag;
    property Top;
    property UseDockManager;
    property Visible;
    property Width;

    property OnActiveControlChange;
    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnDblClick;
    property OnDockDrop;
    property OnDockOver;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
  end;

  TTileMethod = (tmCenter, tmTile);
  TCustomDIBImageContainer = class(TCustomDIBContainer)
  private
    FIndexImage     : TDIBImageLink;
    FTileMethod     : TTileMethod;

    procedure DoImageChanged(Sender : TObject; ID : Integer; Operation : TDIBOperation);
    function  GetDIBImageList: TCustomDIBImageList;
    procedure SetDIBImageList(const Value: TCustomDIBImageList);
    procedure SetTileMethod(const Value: TTileMethod);
  protected
    procedure ImageChanged(ID : Integer; Operation : TDIBOperation); virtual;
    procedure WndProc(var Message : TMessage); override;

    property DIBImageList     : TCustomDIBImageList
      read GetDIBImageList
      write SetDIBImageList;
    property IndexImage       : TDIBImageLink
      read FIndexImage
      write FIndexImage;
    property TileMethod       : TTileMethod
      read FTileMethod
      write SetTileMethod;
  public
    constructor Create(AOwner : TComponent); override;
    destructor destroy; override;

    procedure Paint; override;
  published
  end;

  TDIBImageContainer = class(TCustomDIBImageContainer)
  published
    //New events / properties
    property DIBImageList;
    property IndexImage;
    property TileMethod;

    property OnActiveControlChange;

    //inherited
    property AfterPaint;
    property Align;
    property Anchors;
    property AutoSize;
    property BeforePaint;
    property BiDiMode;
    property BorderDrawPosition;
    property Color;
    property Constraints;
    property Cursor;
    property DockSite;
    property DIBBorder;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property Height;
    property HelpContext;
    property Hint;
    property Left;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property ParentBiDiMode;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property Tag;
    property Top;
    property UseDockManager;
    property Visible;
    property Width;

    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnDblClick;
    property OnDockDrop;
    property OnDockOver;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;

  end;

implementation
uses
  cDIBControl;

type
  EDIBContainerError = class(Exception);
  THackControl = Class(TControl);
  THackWinControl = class(TWinControl);
  THackDIBControl = Class(TCustomDIBControl);
  TBlitType = (btNormal, btNeedPalette, btLookup);

{ TCustomDIBContainer }

procedure TCustomDIBContainer.DoAfterPaint;
begin
  if BorderDrawPosition = bdOverControls then
    if DIBBorder <> nil then DIBBorder.DrawTo(DIB, ClientRect);
  if Assigned(FAfterPaint) then FAfterPaint(Self, FDIB);
end;

procedure TCustomDIBContainer.AlignControls(AControl: TControl;
  var Rect: TRect);
begin
  if DIBBorder <> nil then with DIBBorder do
  begin
    Rect.Top := Rect.Top + BorderTop.Size;
    Rect.Bottom := Rect.Bottom - BorderBottom.Size;
    Rect.Left := Rect.Left + BorderLeft.Size;
    Rect.Right := Rect.Right - BorderRight.Size;
  end;
  inherited;
end;

procedure TCustomDIBContainer.DoBeforePaint;
begin
  if Assigned(FBeforePaint) then FBeforePaint(Self, FDIB);
end;

function TCustomDIBContainer.ChildWantsKey(var Message: TWMKey): Boolean;
var
  DlgResult                   : Integer;
  Mask                        : Integer;
begin
  Result := False;
  if FActiveControl <> nil then with THackDIBControl(FActiveControl) do
  begin
    Message.Result := 1;
    if IsMenuKey(Message) then Exit;
    if not (csDesigning in ComponentState) then with Message do
    begin
      if Perform(CM_CHILDKEY, CharCode, Integer(Self)) <> 0 then exit;
      Mask := 0;
      case CharCode of
        VK_TAB:
          Mask := DLGC_WANTTAB;
        VK_LEFT, VK_RIGHT, VK_UP, VK_DOWN:
          Mask := DLGC_WANTARROWS;
        VK_RETURN, VK_EXECUTE, VK_ESCAPE, VK_CANCEL:
          Mask := DLGC_WANTALLKEYS;
      end;
      if Mask <> 0 then begin
        DlgResult := Perform(WM_GETDLGCODE, 0, 0);
        if DlgResult <> DLGC_WantAllKeys then begin
          if Mask and DlgResult = 0 then exit
        end;
      end;
    end;
    Result := True;
  end;
end;

procedure TCustomDIBContainer.CMFocusChanged(var Message: TCMFocusChanged);
var
  KS                          : TKeyboardState;
begin
  if Message.Sender <> nil then
    if TWinControl(Message.Sender).Parent = Self then begin
      DIBFocusControl(Message.Sender);
      exit;
    end;

  if (TWinControl(Message.Sender) = Self) then begin
    if DIBGetLast is TWincontrol then begin
      GetKeyboardState(KS);
      if KS[VK_SHIFT] and 128 = 128 then DIBSelectPrior;
    end;
  end else
    DIBFocusControl(nil);
end;

procedure TCustomDIBContainer.CNChar(var Message: TWMKey);
begin
  if ChildWantsKey(Message) then begin
    Message.Result := FActiveControl.Perform(WM_Char,TMessage(Message).WParam,
      TMessage(Message).LParam);
    exit;
  end;
end;

procedure TCustomDIBContainer.CNKeyDown(var Message: TWMKeyDown);
var
  KS                          : TKeyboardState;
  Handled                     : Boolean;
begin
  if ChildWantsKey(Message) then begin
    Message.Result := FActiveControl.Perform(WM_KeyDown,TMessage(Message).WParam,
      TMessage(Message).LParam);
    exit;
  end;

  if (Screen.ActiveControl = DIBGetLast) then
  begin
    inherited;
    exit;
  end;

  Handled := False;
  if Message.CharCode = VK_TAB then
  try
    FTabbing := True;
    GetKeyboardState(KS);
    if KS[VK_SHIFT] and 128 = 128 then
      Handled := DIBSelectPrior
    else
      Handled := DIBSelectNext
  finally
    FTabbing := False;
  end;

  if not Handled then begin
    inherited;
    if Screen.ActiveControl = Self then DoEnter;
  end;
end;

procedure TCustomDIBContainer.CNKeyUp(var Message: TWMKeyUp);
begin
  if ChildWantsKey(Message) then begin
    Message.Result := FActiveControl.Perform(WM_KeyUp,TMessage(Message).WParam,
      TMessage(Message).LParam);
    exit;
  end;
end;

constructor TCustomDIBContainer.Create(AOwner: TComponent);
begin
  inherited;
  FTabList := TList.Create;
  ControlStyle := ControlStyle + [csAcceptsControls];
  FDIB := TWinDib.Create;
  FChildDIB := TWinDIB.Create;
  DoubleBuffered := True;
  TabStop := False;
  FActiveControl := nil;
  FChangingFocus := False;
  FTabbing := False;
  BorderWidth := 0;
  BorderStyle := bsNone;
  BevelOuter := bvNone;
  FPalette := nil;
  FAlteredRect := False;
  BorderStyle := bsNone;
end;

procedure TCustomDIBContainer.CreateParams(var Params: TCreateParams);
begin
  inherited;
end;

destructor TCustomDIBContainer.Destroy;
begin
  FTabList.Free;
  FChildDIB.Free;
  FDIB.Free;
  inherited;
end;

function TCustomDIBContainer.DIBFindNext(aControl: TControl;
  aForward: Boolean): TControl;
var
  I, StartIndex: Integer;
begin
  Result := nil;

  if FTabList.Count > 0 then begin
    StartIndex := FTabList.IndexOf(aControl);
    if StartIndex = -1 then
      if aForward then
        StartIndex := FTabList.Count - 1
      else
        StartIndex := 0;

    I := StartIndex;
    repeat
      if aForward then begin
        Inc(I);
        if I = FTabList.Count then I := 0;
      end else begin
        if I = 0 then I := FTabList.Count;
        Dec(I);
      end;

      aControl := FTabList[I];
      if aControl.Enabled then Result := aControl;
    until (Result <> nil) or (I = StartIndex);
  end;
end;

procedure TCustomDIBContainer.DIBFocusControl(aControl: TControl);
begin
  if FChangingFocus then exit;
  FChangingFocus := True;
  try
    if aControl <> FActiveControl then
      if FActiveControl <> nil then
        FActiveControl.Perform(WM_KillFocus,0,0);

    FActiveControl := aControl;
    if aControl <> nil then FActiveControl.Perform(WM_SetFocus,0,0);
    if Assigned(FOnActiveControlChange) then
      FOnActiveControlChange(Self);
  finally
    FChangingFocus := False;
  end;
end;

function TCustomDIBContainer.DIBGetFirst: TControl;
begin
  Result := nil;
  if FTabList.Count > 0 then Result := TControl(FTabList[0]);

  if Result = nil then begin
    Result := FindNextControl(Self,True,True,True);
    if Result <> nil then
      if Result.Parent <> self then
        Result := nil;
  end;
end;

function TCustomDIBContainer.DIBGetIndex(aControl: TControl): Integer;
begin
  Result := -1;
  if aControl is TWinControl then begin
    if aControl.Parent = Self then
      Result := TWinControl(aControl).TabOrder + ControlCount;
  end else begin
    Result := FTabList.IndexOf(aControl);
  end;
end;

function TCustomDIBContainer.DIBGetLast: TControl;
var
  OrderList                   : TList;
begin
  Result := nil;
  OrderList := TList.Create;
  try
    GetTabOrderList(OrderList);
    if OrderList.Count > 0 then
      Result := OrderList[OrderList.Count-1]
    else
      if FTabList.Count > 0 then
        Result := FTabList[FTabList.Count-1];
  finally
    OrderList.Free;
  end;
end;

function TCustomDIBContainer.DIBGetNext: TControl;
begin
  Result := nil;
  if not (FActiveControl is TWinControl) then begin
    Result := DIBFindNext(FActiveControl,True);
    if DIBGetIndex(Result) < DIBGetIndex(FActiveControl) then Result := nil;
  end;

  if (FActiveControl is TWinControl) or (Result=nil) then begin
    if Result = nil then
      Result := FindNextControl(Self,True,True,True)
    else
      Result := FindNextControl(TWinControl(FActiveControl),True,True,True);
    if Result <> nil then
      if (Result.Parent <> self) or (DIBGetIndex(Result) < DIBGetIndex(FActiveControl)) then
        Result := nil;
  end;

  if DIBGetIndex(Result) < DIBGetIndex(FActiveControl) then Result := nil;
end;

function TCustomDIBContainer.DIBGetPrior: TControl;
begin
  Result := nil;
  if (FActiveControl is TWinControl) or (Result=nil) then begin
    //Look backwards
    Result := FindNextControl(TWinControl(FActiveControl),False,True,True);
    if Result <> nil then
      if (Result.Parent <> self) then Result := nil;
  end;

  if (DIBGetIndex(Result) > DIBGetIndex(FActiveControl)) or
     (DIBGetIndex(Result) =-1)
  then Result := nil;

  if not (FActiveControl is TWinControl) or (Result=nil) then begin
    //Look backwards
    Result := DIBFindNext(FActiveControl, False);
    if (DIBGetIndex(Result) > DIBGetIndex(FActiveControl)) or
       (DIBGetIndex(Result) =-1)
    then Result := nil;
  end;

  if (DIBGetIndex(Result) > DIBGetIndex(FActiveControl)) or
     (DIBGetIndex(Result) =-1)
  then Result := nil;
end;

function TCustomDIBContainer.DIBGetTabOrder(aControl: TControl): TTabOrder;
begin
  Result := FTabList.IndexOf(aControl);
end;

function TCustomDIBContainer.DIBSelectNext: Boolean;
var
  NewControl,
  OldControl                  : TControl;
begin
  Result := False;
  OldControl := FActiveControl;

  NewControl := DIBGetNext;


  if NewControl <> nil then
    if DIBGetIndex(NewControl) > DIBGetIndex(OldControl) then
      Result := True;

  ActiveControl := NewControl;
end;

function TCustomDIBContainer.DIBSelectPrior: Boolean;
var
  NewControl,
  OldControl                  : TControl;
begin
  Result := False;
  OldControl := FActiveControl;

  NewControl := DIBGetPrior;


  if NewControl <> nil then
    if DIBGetIndex(NewControl) < DIBGetIndex(OldControl) then
      if DIBGetIndex(NewControl) <> -1 then
        Result := True;

  ActiveControl := NewControl;
end;

procedure TCustomDIBContainer.DIBSetTabOrder(aControl: TControl;
  NewIndex: TTabOrder);
var
  OldIndex                    : TTabOrder;
begin
  if aControl = nil then exit;
  OldIndex := FTabList.IndexOf(aControl);
  if OldIndex = NewIndex then exit;

  if (OldIndex >=0) then FTabList.Delete(OldIndex);
  if NewIndex >= FTabList.Count then
    FTabList.Add(aControl)
  else
    if NewIndex >=0 then
      FTabList.Insert(NewIndex, aControl);
  TabStop := FTabList.Count > 0;
end;

procedure TCustomDIBContainer.DoEnter;
var
  KS                          : TKeyboardState;
  NextControl                 : TControl;
begin
  if FActiveControl = nil then
    if FTabList.Count > 0 then begin
      GetKeyboardState(KS);
      if (KS[VK_SHIFT] and 128) = 128 then
        NextControl := DIBGetLast
      else
        NextControl := DIBGetFirst;

      DIBFocusControl(NextControl);
      if Assigned(OnEnter) then OnEnter(Self);
    end else
      inherited;
end;

procedure TCustomDIBContainer.DoExit;
begin
  DIBFocusControl(nil);
  inherited;
end;

procedure TCustomDIBContainer.DoTabFixups;
var
  List                        : Array of Pointer;
  CC,
  X                           : Integer;
begin
  CC := ControlCount;
  SetLength(List,CC);
  for X:=0 to CC-1 do List[X] := nil;
  for X:=0 to CC-1 do
    if Controls[X] is TCustomDIBControl then
      with THackDIBControl(Controls[X]) do
        if FTabOrder <> -1 then
          if FTabOrder < CC then
            List[FTabOrder] := Controls[X]
          else begin
            CC := FTabOrder+1;
            SetLength(List,FTabOrder);
            List[FTabOrder] := Controls[X];
          end;

  for X:=0 to CC -1 do
    if List[X] <> nil then DIBSetTabOrder(List[X],X);
end;

procedure TCustomDIBContainer.Loaded;
begin
  inherited;
  FDIB.Resize(Width,Height);
  DoTabFixups;
end;

procedure TCustomDIBContainer.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if (AComponent = self) or (csDestroying in ComponentState) then exit;

  if aComponent is TDIBPalette then begin
    if Operation = opInsert then
      FPalette := TDIBPalette(AComponent);
    if (Operation = opRemove) and (FPalette = AComponent) then
      FPalette := Nil;
  end;
  if (Operation = opRemove) then
  begin
    if AComponent is TControl then DIBSetTabOrder(TControl(aComponent),-1);
    if AComponent = DIBBorder then DIBBorder := nil;
  end;
end;

procedure TCustomDIBContainer.Paint;
begin
  if (FUpdateRect.Left = 0) and (FUpdateRect.Top=0) and
     (FUpdateRect.Right = Width) and (FUpdateRect.Bottom = height) then
    FDIB.QuickFill(ColorToRGB(Self.Color))
  else
    with FUpdateRect do
      FDIB.QuickFillRect(ColorToRGB(Self.Color), Left, Top, Right-Left, Bottom-Top);
end;

procedure TCustomDIBContainer.PaintControls(DC: HDC; First: TControl);
var
  I, Count, SaveIndex: Integer;
  FrameBrush: HBRUSH;
  D : TRect;
begin
  if BorderDrawPosition = bdUnderControls then
    if DIBBorder <> nil then DIBBorder.DrawTo(DIB, ClientRect);
  if DockSite and UseDockManager and (DockManager <> nil) then
    DockManager.PaintSite(DC);
  if ControlCount > 0 then
  begin
    //Do non-wincontrols
    I := 0;
    if First <> nil then for Count := ControlCount downto 0 do
      if Controls[I] = First then begin
        I := Count;
        Break;
      end;
    Count := ControlCount;
    while I < Count do begin
      if IntersectRect(D,FUpdateRect,Controls[I].BoundsRect)  then
      if not (Controls[I] is TWinControl) then
      begin
        with TControl(Controls[I]) do
          if (Visible or (csDesigning in ComponentState) and
            not (csNoDesignVisible in ControlStyle)) and
            RectVisible(DC, Rect(Left, Top, Left + Width, Top + Height)) then
          begin
            if csPaintCopy in Self.ControlState then ControlState := ControlState + [csPaintCopy];
            SaveIndex := SaveDC(DC);
            MoveWindowOrg(DC, Left, Top);
            IntersectClipRect(DC, 0, 0, Width, Height);

            if Controls[I] is TCustomDIBControl then
            begin
              FChildDIB.ReSize(Width,Height);
              FChildDIB.ResetHeader;
              THackDIBControl(Controls[I]).ControlDIB := FChildDIB;
            end;

            Perform(WM_PAINT, DC, 0);
            RestoreDC(DC, SaveIndex);
            ControlState := ControlState - [csPaintCopy];
          end;
      end;
      Inc(I);
    end;

    //Now do WinControls (last)
    I := 0;
    if First <> nil then for Count := ControlCount downto 0 do
      if Controls[I] = First then begin
        I := Count;
        Break;
      end;
    Count := ControlCount;
    while I < Count do begin
      if IntersectRect(D,FUpdateRect,Controls[I].BoundsRect)  then
      if (Controls[I] is TWinControl) then begin
        with THackWinControl(Controls[I]) do begin
          if Ctl3D and (csFramed in ControlStyle) and
            (Visible or (csDesigning in ComponentState) and
            not (csNoDesignVisible in ControlStyle)) then
          begin
            FrameBrush := CreateSolidBrush(ColorToRGB(clBtnShadow));
            FrameRect(DC, Rect(Left - 1, Top - 1, Left + Width, Top + Height),
              FrameBrush);
            DeleteObject(FrameBrush);
            FrameBrush := CreateSolidBrush(ColorToRGB(clBtnHighlight));
            FrameRect(DC, Rect(Left, Top, Left + Width + 1, Top + Height + 1),
              FrameBrush);
            DeleteObject(FrameBrush);
          end;
        end;
      end;
      Inc(I);
    end;
  end;
end;

procedure TCustomDIBContainer.PaintHandler(var Message: TWMPaint);
var
  I, Clip, SaveIndex: Integer;
  DC: HDC;
begin
  DC := Message.DC;
  if ControlCount = 0 then
    PaintWindow(DC)
  else begin
    SaveIndex := SaveDC(DC);
    Clip := SimpleRegion;
    for I := 0 to ControlCount - 1 do
      with TControl(Controls[I]) do
        if (Visible or (csDesigning in ComponentState) and
          not (csNoDesignVisible in ControlStyle)) and
          (csOpaque in ControlStyle) then
        begin
          Clip := ExcludeClipRect(DC, Left, Top, Left + Width, Top + Height);
          if Clip = NullRegion then Break;
        end;
    if Clip <> NullRegion then PaintWindow(DC);
    RestoreDC(DC, SaveIndex);
  end;
  PaintControls(DC, nil);
end;

procedure TCustomDIBContainer.SetBorderDrawPosition(
  const Value: TBorderDrawPosition);
begin
  FBorderDrawPosition := Value;
  Invalidate;
end;

procedure TCustomDIBContainer.SetBounds(aLeft, aTop, aWidth, aHeight: Integer);
begin
  inherited;
  if (FDIB <> nil) and not (csLoading in ComponentState) then FDIB.Resize(aWidth,aHeight);
end;

procedure TCustomDIBContainer.SetDIBBorder(const Value: TDIBBorder);
begin
  if DIBBorder <> nil then DIBBorder.RemoveFreeNotification(Self);
  FDIBBorder := Value;
  if DIBBorder <> nil then DIBBorder.FreeNotification(Self);
  Invalidate;
end;

procedure TCustomDIBContainer.WMEraseBkGnd(var Message: TMessage);
begin
  
//  inherited;
end;

procedure TCustomDIBContainer.WMGetDlgCode(var Message: TMessage);
begin
  Message.Result := DLGC_WANTALLKEYS;
end;

procedure TCustomDIBContainer.WMPaint(var Message: TWMPaint);
var
  I                 : Integer;
  OrigDC,
  DC                : HDC;
  PS                : TPaintStruct;
  OldPal            : HPalette;
  BlitType          : TBlitType;
  CanSetDiBits      : Boolean;
begin
  DC := message.DC;
  if Message.DC = 0 then dc := beginpaint(handle,ps);
  try
    // painting to the control (the norm)
    if Message.DC = 0 then
      FUpdateRect := PS.rcPaint        // get the area we will be painting in
    // for painting to an alternate DC (non-owned canvas)
    else begin
      FAlteredRect:=True;                    // stop height/width, etc checks
      GetClipBox(DC,FUpdateRect);    // get the area we will be painting in
    end;

    if not FAlteredRect then begin
      //When another form passes over our form, the display screws up,
      //if we ALWAYS paint the whole image, we have no problems !
      if GetParentForm(Self) <> nil then
        if GetActiveWindow <> GetParentForm(Self).handle then FUpdateRect := ClientRect;

//      if csDesigning in ComponentState then
//        FUpdateRect := ClientRect;

      for I := 0 to ControlCount -1 do
        if Controls[I] is TCustomDIBControl then
          THackDIBControl(Controls[I]).AlterUpdateRect(FUpdateRect);

      FUpdateRect.TopLeft := ClientToScreen(FUpdateRect.TopLeft);
      FUpdateRect.BottomRight := ClientToScreen(FUpdateRect.BottomRight);

      if FUpdateRect.Left < 0 then FUpdateRect.Left := 0;
      if FUpdateRect.Top < 0 then FUpdateRect.Top := 0;
      if FUpdateRect.Right > Screen.Width then FUpdateRect.Right := Screen.Width;
      if FUpdateRect.Bottom > Screen.Height then FUpdateRect.Bottom := Screen.Height;

      FUpdateRect.TopLeft := ScreenToClient(FUpdateRect.TopLeft);
      FUpdateRect.BottomRight := ScreenToClient(FUpdateRect.BottomRight);

      if FUpdateRect.Left < 0 then FUpdateRect.Left := 0;
      if FUpdateRect.Top < 0 then FUpdateRect.Top := 0;
      if FUpdateRect.Right > ClientWidth then FUpdateRect.Right := ClientWidth;
      if FUpdateRect.Bottom > ClientHeight then FUpdateRect.Bottom := ClientHeight;


      //For some reason, invalidating the whole form actually invalidates the height -2
      //so the next code is called endlessly.
      //Therefore I check if the difference in height is > 2
      if (FUpdateRect.Left <> PS.rcPaint.Left) or
         (FUpdateRect.Top <> PS.rcPaint.Top) or
         (FUpdateRect.Right <> PS.rcPaint.Right) or
         (abs(FUpdateRect.Bottom - PS.rcPaint.Bottom) > 2) then
  //       (FUpdateRect.Bottom <> PS.rcPaint.Bottom) then
      begin
        FAlteredRect := True;
        if Message.DC = 0 then endpaint(handle, ps);
        ValidateRect(Handle, @PS.rcPaint);
        InvalidateRect(Handle, @FUpdateRect, False);
        exit;
      end;
    end;
    FAlteredRect := False;

    FDIB.ClipRect := FUpdateRect;
    OrigDC := Message.DC;
    Message.DC := DIB.Handle;
    DoBeforePaint;
    PaintHandler(Message);
    DoAfterPaint;
    Message.DC := OrigDC;

    BlitType := btNormal;
    if not (csDesigning in ComponentState) then
    if GetDeviceCaps(DC, BITSPixel) = 8 then begin
      CanSetDiBits := (GetDeviceCaps(DC, RasterCaps) and RC_DIBToDEV) <> 0;
      if (FPalette <> nil) and (FPalette.UseTable) and (CanSetDiBits) then
        BlitType := btLookUp
      else
        BlitType := btNeedPalette
    end;

    case BlitType of
      btNormal :
        with FUpdateRect do
          BitBlt(DC, Left, Top, Right-Left, Bottom-Top , dib.handle, Left, Top, SrcCopy);

      btNeedPalette :
        begin
          OldPal := SelectPalette(DC, FPalette.palette, FALSE);
          with FUpdateRect do
            BitBlt(DC, Left, Top, Right-Left, Bottom-Top , dib.handle, Left, Top, SrcCopy);
          SelectPalette(DC, OldPal, TRUE);
        end;

      btLookup:
        begin
          OldPal := SelectPalette(DC, FPalette.palette, FALSE);
          with FUpdateRect do
            dib.Render8Bit(DC, Left, Top, Right-Left, Bottom-Top, Left, Top, SrcCopy, FPalette);
          SelectPalette(DC, OldPal, TRUE);
        end;
    end;
  finally
    if Message.DC = 0 then EndPaint(Handle,PS);
  end;
  DefaultHandler(Message);
end;


procedure TCustomDIBContainer.WndProc(var Message: TMessage);
var
  CControl          : TControl;
  ParentRect,
  CRect             : TRect;
  CPos              : TPoint;
  ParentControl     : TWinControl;
begin
  if Message.Msg = WM_LButtonUp then begin
    CControl := GetCaptureControl;
    if (CControl <> nil) then begin
      Windows.GetCursorPos(CPos);
      CRect.TopLeft := CControl.ClientToScreen(Point(CControl.Left,CControl.Top));
      CRect.BottomRight := Point(CRect.Left + CControl.Width, CRect.Top + CControl.Height);
      ParentRect := CRect;
      if not PtInRect(CRect, CPos) then begin
        if THackControl(CControl).MouseCapture then
          if CControl <> Self then CControl.Perform(CM_MouseLeave,0,0);

        ParentControl := FindVCLWindow(CPos);
        if ParentControl <> nil then begin
          CRect.TopLeft := ParentControl.ScreenToClient(CPos);
          CControl := ParentControl.ControlAtPos(CRect.TopLeft,False);
          if CControl <> nil then CControl.Perform(CM_MouseEnter,0,0);
        end;
      end;
    end;
  end;

(*
  if FActiveControl is TCustomDIBControl then begin
    if not FindDIBChildMessage(Message.Msg) then
      inherited
    else begin
      FActiveControl.WindowProc(Message);
      if Message.Result <> 0 then
        inherited;
    end;
  end else
    inherited;
*)
  inherited;
end;

{ TCustomDIBImageContainer }

constructor TCustomDIBImageContainer.Create(AOwner: TComponent);
begin
  inherited;
  FIndexImage := TDIBImageLink.Create(Self);
  FIndexImage.OnImageChanged := DoImageChanged;
end;

destructor TCustomDIBImageContainer.destroy;
begin
  FIndexImage.Free;
  inherited;
end;

procedure TCustomDIBImageContainer.Paint;
var
  X,
  Y                 : Integer;
  R                 : TRect;
  TheDIB            : TMemoryDIB;
begin
  if not FIndexImage.GetImage(TheDIB) then begin
    inherited;
    exit;
  end;

  if TileMethod <> tmTile then
    if (TheDIB.Width <> Width) or (TheDIB.Height <> Height) then
      Inherited;

  if TheDIB.Height > 0 then begin
    case TileMethod of
      tmCenter  :
        begin
          TheDIB.Draw(Width div 2 - (TheDIB.Width div 2),Height div 2 - (TheDIB.Height div 2),
            TheDIB.Width,TheDIB.Height,FDIB,0,0);
        end;
      tmTile    :
        begin
          Y := 0;
          while Y < Height do begin
            X := 0;
            while X < Width do begin
              if IntersectRect(R,FUpdateRect,Rect(X,Y,X+TheDIB.Width,Y+TheDIB.Height)) then
                TheDIB.Draw(X,Y,TheDIB.Width,TheDIB.Height,FDIB,0,0);
              Inc(X, TheDIB.Width);
            end;
            Inc(Y, TheDIB.Height);
          end;
        end;
    end;
  end;
end;

procedure TCustomDIBImageContainer.SetTileMethod(const Value: TTileMethod);
begin
  FTileMethod := Value;
  invalidate;
end;

procedure TCustomDIBImageContainer.WndProc(var Message: TMessage);
begin
  if (csDestroying in ComponentState) or
     (TileMethod <> tmTile) or
     (Message.msg <> WM_EraseBkGnd) then
    inherited;
end;

procedure TCustomDIBImageContainer.ImageChanged(ID: Integer; Operation: TDIBOperation);
begin
  case Operation of
    doRemove :
      if ID = IndexImage.DIBIndex then
        IndexImage.DIBIndex := -1
      else
      if ID < IndexImage.DIBIndex then
        IndexImage.DIBIndex := IndexImage.DIBIndex -1;
    doChange : if ID = IndexImage.DIBIndex then Invalidate;
  end;
end;

function TCustomDIBImageContainer.GetDIBImageList: TCustomDIBImageList;
begin
  Result := FIndexImage.DIBImageList;
end;

procedure TCustomDIBImageContainer.SetDIBImageList(const Value: TCustomDIBImageList);
begin
  FIndexImage.DIBImageList := Value;
end;

procedure TCustomDIBImageContainer.DoImageChanged(Sender: TObject;
  ID: Integer; Operation: TDIBOperation);
begin
  ImageChanged(ID, Operation);
end;

end.
